{******************************************}
{                                          }
{             FastReport v4.0              }
{  IB Objects (Native) enduser components  }
{             IB Objects 4.7               }
{  Version-Date: 2007-02-01                }
{                                          }
{ Created by: Andreas Hesse                }
{ E-mail: delphi@ah3000.de                 }
{                                          }
{******************************************}

unit frxIB_Components;

interface

{$I frx.inc}
{$R frxIB_Components.res}

uses
  Windows,
  Classes,
  frxClass,
  frxIBOSet,
  frxrcIB_,
  IB_Components
{$IFDEF Delphi6}
, Variants
{$ENDIF}
{$IFDEF QBUILDER}
//, fqbClass
{$ENDIF};

type

  (** IB Component for FastReport *)
  TfrxIB_Components = class(TfrxDBComponents)
  private
    FOldComponents: TfrxIB_Components;
    FDefaultConnection: TIB_Connection;
    FDefaultTransaction: TIB_Transaction;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetDescription: String; override;
  published
    property DefaultConnection: TIB_Connection read FDefaultConnection write FDefaultConnection;
    property DefaultTransaction: TIB_Transaction read FDefaultTransaction write FDefaultTransaction;
  end;

  TfrxIB_Database = class(TfrxCustomDatabase)
  private
    FDatabase: TIB_Connection;
    FTransaction: TIB_Transaction;
    function GetAbout: string;
  protected
    function GetConnected: Boolean; override;
    function GetDatabaseName: String; override;
    function GetLoginPrompt: Boolean; override;
    function GetParams: TStrings; override;
    procedure SetConnected(Value: Boolean); override;
    procedure SetDatabaseName(const Value: String); override;
    procedure SetLoginPrompt(Value: Boolean); override;
    procedure SetParams(Value: TStrings); override;
    procedure SetTransaction(Value: TIB_Transaction);
  public
    class function GetDescription: String; override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // procedure SetLogin(const Login, Password: String); override;
    property Database: TIB_Connection read FDatabase;
  published
    property DatabaseName: String read GetDatabaseName write SetDatabaseName;
    property LoginPrompt: Boolean read GetLoginPrompt write SetLoginPrompt default True;
    property Params: TStrings read GetParams write SetParams;
    // property SQLDialect: Integer read GetSQLDialect write SetSQLDialect;
    property Connected: Boolean read GetConnected write SetConnected default False;
    property Version :string read GetAbout stored False;
  end;

  (** @see TfrxParamItem *)
  TfrxIB_ParamItem = class(TCollectionItem)
  private
    FExpression: String;
    FName: String;
    FValue: Variant;
  public
    procedure Assign(Source: TPersistent); override;
    property Value: Variant read FValue write FValue;
  published
    property Name: String read FName write FName;
    property Expression: String read FExpression write FExpression;
  end;

  (** @see TfrxParams *)
  TfrxIB_Params = class(TCollection)
  private
    function GetParam(Index: Integer): TfrxIB_ParamItem;
  public
    constructor Create;
    function Add: TfrxIB_ParamItem;
    function Remove(const Name: string): boolean;
    function PutParam(const Name, Expression: String): TfrxIB_ParamItem;
    function Find(const Name: String): TfrxIB_ParamItem;
    function IndexOf(const Name: String): Integer;
    property Items[Index: Integer]: TfrxIB_ParamItem read GetParam; default;
  end;

  (** native IBO query for Designer *)
  TfrxIB_Query = class(TfrxIBODataset)
  { TfrxIB_Query }
  private
    FQuery: TIB_Query;
    FDatabase: TfrxIB_Database;
    function GetDatabase: TfrxIB_Database;
    procedure SetDatabase(const Value: TfrxIB_Database);
    function GetMasterLinks: TStrings;
    procedure SetMasterLinks(const Value: TStrings);
    function GetMasterParamLinks: TStrings;
    procedure SetMasterParamLinks(const Value: TStrings);
    function GetPrepared: boolean;
    procedure SetPrepared(const Value: boolean);
    function GetVersion: string;
    procedure SetVersion(const Value: string);
    function GetTransaction: TIB_Transaction;
    procedure SetTransaction(const Value: TIB_Transaction);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoAfterPrepare(Sender: TIB_Statement); virtual;
    procedure DoBeforeOpen(IB_Dataset: TIB_Dataset ); virtual;
    procedure SetParent(AParent: TfrxComponent); override;
    procedure SetUserName(const Value: String); override;
  public
    class function GetDescription: String; override;
    constructor Create(AOwner: TComponent); override;
    constructor DesignCreate(AOwner: TComponent; Flags: Word); override;
    destructor Destroy; override;
    procedure UpdateDataTree;
    property Query: TIB_Query read FQuery;
  published
    property Database: TfrxIB_Database read GetDatabase write SetDatabase;
    property MasterLinks: TStrings read GetMasterLinks write SetMasterLinks;
    property MasterParamLinks: TStrings read GetMasterParamLinks write SetMasterParamLinks;
    property Prepared: boolean read GetPrepared write SetPrepared stored false;
    property Transaction: TIB_Transaction read GetTransaction write SetTransaction stored false;
    property Version: string read GetVersion write SetVersion stored false;
  { TfrxCustomQuery }
  private
    FParams: TfrxIB_Params;
    procedure ReadData(Reader: TReader);
    procedure SetParams(Value: TfrxIB_Params);
    procedure WriteData(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure SetSQL(Value: TStrings); virtual;
    function GetSQL: TStrings; virtual;
  public
    procedure UpdateParams; virtual;
    function ParamByName(const Value: String): TfrxIB_ParamItem;
  published
    property Params: TfrxIB_Params read FParams write SetParams;
    property SQL: TStrings read GetSQL write SetSQL;
  { TfrxCustomDataset }
  private
    FDataSource: TIB_DataSource;
    FMaster: TfrxIBODataset;
    procedure SetActive(Value: Boolean);
    procedure SetFilter(const Value: String);
    procedure SetFiltered(Value: Boolean);
    function GetActive: Boolean;
    function GetFields: TIB_Row;
    function GetFilter: String;
    function GetFiltered: Boolean;
    procedure InternalSetMaster(const Value: TfrxIBODataset);
  public
    property Fields: TIB_Row read GetFields;
    property Active: Boolean read GetActive write SetActive default False;
  published
    property Filter: String read GetFilter write SetFilter;
    property Filtered: Boolean read GetFiltered write SetFiltered default False;
    property Master: TfrxIBODataset read FMaster write InternalSetMaster;
  end;

var
  GlobalIB_Components: TfrxIB_Components;

const
  (** Versionnumber of Component (Version of FR) *)
  S_VERSION = '1.1.1 (FR 4.01)';

implementation

uses
  SysUtils,
  Graphics,
  frxIB_RTTI,
  frxUtils,
  frxRes,
{$IFNDEF NO_EDITORS}
  frxIB_Editor,
{$ENDIF}
  frxDsgnIntf;

{ TfrxIB_Components }

constructor TfrxIB_Components.Create(AOwner: TComponent);
begin
  inherited;
  FOldComponents:= GlobalIB_Components;
  GlobalIB_Components := Self;
end;

destructor TfrxIB_Components.Destroy;
begin
  if GlobalIB_Components = Self then
    GlobalIB_Components := FOldComponents;
  inherited;
end;

function TfrxIB_Components.GetDescription: String;
begin
  Result := 'IB Objects';
end;

{ TfrxIB_Connection }

class function TfrxIB_Database.GetDescription: String;
begin
  Result := 'IB Database';
end;

constructor TfrxIB_Database.Create(AOwner: TComponent);
begin
  FDatabase := TIB_Connection.Create(nil);  // owner = nil to avoid double freeing
  SetTransaction(nil);  // set default transaction
  LoginPrompt := true;  // default value
  Component := FDatabase;
  inherited;
end;

(** FConnection will be free'd with Component in the BaseClass *)
destructor TfrxIB_Database.Destroy;
begin
  FTransaction.Free;
  inherited;
end;

function TfrxIB_Database.GetConnected: Boolean;
begin
  Result := FDatabase.Connected;
end;

function TfrxIB_Database.GetDatabaseName: String;
begin
  Result := FDatabase.DatabaseName;
end;

function TfrxIB_Database.GetLoginPrompt: Boolean;
begin
  Result := FDatabase.LoginPrompt;
end;

function TfrxIB_Database.GetParams: TStrings;
begin
  Result := FDatabase.Params;
end;

procedure TfrxIB_Database.SetConnected(Value: Boolean);
begin
  BeforeConnect(Value);
  FDatabase.Connected := Value;
end;

procedure TfrxIB_Database.SetDatabaseName(const Value: String);
begin
  FDatabase.DatabaseName := Value;
end;

procedure TfrxIB_Database.SetLoginPrompt(Value: Boolean);
begin
  FDatabase.LoginPrompt := Value;
end;

procedure TfrxIB_Database.SetParams(Value: TStrings);
begin
  FDatabase.Params.Assign(Value);
end;

procedure TfrxIB_Database.SetTransaction(Value: TIB_Transaction);
begin
  if Value = nil then
  begin
    if Assigned(GlobalIB_Components) then
    begin
      FDatabase.DefaultTransaction := GlobalIB_Components.DefaultTransaction;
    end
    else
    begin
      if not Assigned(FTransaction) then
      begin
        FTransaction := TIB_Transaction.create(nil);
        FTransaction.ReadOnly := true;
        FTransaction.Isolation := tiCommitted;
        FTransaction.IB_Connection := FDatabase;
      end;
      FDatabase.DefaultTransaction := FTransaction;
    end;
    FDatabase.DefaultTransaction.AddConnection(FDatabase);
  end;
end;

function TfrxIB_Database.GetAbout: string;
begin
  Result := S_VERSION;
end;

{ TfrxIB_ParamItem }

procedure TfrxIB_ParamItem.Assign(Source: TPersistent);
begin
  if Source is TfrxIB_ParamItem then
  begin
    FName := TfrxIB_ParamItem(Source).Name;
    FExpression := TfrxIB_ParamItem(Source).Expression;
    FValue := TfrxIB_ParamItem(Source).Value;
  end;
end;

{ TfrxIB_Params }

function TfrxIB_Params.IndexOf(const Name: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to Count - 1 do
    if CompareText(Items[i].Name, Name) = 0 then
    begin
      Result := i;
      break;
    end;
end;

function TfrxIB_Params.GetParam(Index: Integer): TfrxIB_ParamItem;
begin
  Result := TfrxIB_ParamItem(inherited Items[Index]);
end;

constructor TfrxIB_Params.Create;
begin
  inherited Create(TfrxIB_ParamItem);
end;

function TfrxIB_Params.Add: TfrxIB_ParamItem;
begin
  Result := TfrxIB_ParamItem(inherited Add);
end;

function TfrxIB_Params.Find(const Name: String): TfrxIB_ParamItem;
var
  i: Integer;
begin
  i := IndexOf(Name);
  if i <> -1 then
    Result := Items[i] else
    Result := nil;
end;

(**
 * If the Parameter does not exist, it will add a new Parameter to the collection.
 * If the Parameter exist, it will change the expression of this parameter.
 *)
function TfrxIB_Params.PutParam(const Name, Expression: String): TfrxIB_ParamItem;
begin
  Result := Find(Name);
  if Assigned(Result) then begin
    Result.Expression := Expression;
  end
  else
    Result := Add;
    Result.Name := Name;
    Result.Expression := Expression;
  begin
  end;
end;

(** Remove the item with the Parameter Name from the collection *)
function TfrxIB_Params.Remove(const Name: string): boolean;
var i: integer;
begin
  i := IndexOf(Name);
  Result := (i<>-1);
  if Result then begin
    Delete(i);
  end;
end;

{ TfrxIB_Query }

class function TfrxIB_Query.GetDescription: String;
begin
  Result:= 'IBO_Query';
end;

(**
 * @see TfrxCustomDataset.Create
 * @see TfrxCustomQuery.Create
 *)
constructor TfrxIB_Query.Create(AOwner: TComponent);
begin
  FQuery  := TIB_Query.Create(nil);  // owner = nil to avoid double freeing because Component will be freed
  Component := FQuery;
  FQuery.AfterPrepare := DoAfterPrepare;
  FQuery.BeforeOpen := DoBeforeOpen;
  Dataset := FQuery;
  SetDatabase(nil);  // = set default connection
  { TfrxCustomQuery }
  FParams := TfrxIB_Params.Create;
  inherited;
  CloseDataSource := True;
  FDataSource := TIB_DataSource.Create(nil);
  FQuery.MasterSource := FDataSource;  // = SetMaster(..)
end;

constructor TfrxIB_Query.DesignCreate(AOwner: TComponent; Flags: Word);
var
  i: Integer;
  l: TList;
begin
  inherited;
  l := Report.AllObjects;
  for i := 0 to l.Count - 1 do
    if TObject(l[i]) is TfrxIB_Database then
    begin
      SetDatabase(TfrxIB_Database(l[i]));
      break;
    end;
end;

(**
 * @see TfrxCustomDataset.Destroy
 * @see TfrxCustomQuery.Destroy
 *)
destructor TfrxIB_Query.Destroy;
begin
  FParams.Free;
  FDataSource.Free;
  inherited;
end;

function TfrxIB_Query.GetDatabase: TfrxIB_Database;
begin
  Result := FDatabase;
end;

procedure TfrxIB_Query.SetDatabase(const Value: TfrxIB_Database);
begin
  if (Value <> FDatabase) or not Assigned(FQuery.IB_Connection) then
  begin
    FQuery.Unprepare;
    FQuery.IB_Transaction := nil;
    FDatabase := Value;
    if Assigned(Value) then
    begin
      FQuery.IB_Connection := Value.Database;
    end
    else
    begin
      if Assigned(GlobalIB_Components) then
        FQuery.IB_Connection := GlobalIB_Components.DefaultConnection
      else
        FQuery.IB_Connection := nil;
    end;
    if Assigned(FQuery.IB_Connection) then
      FQuery.CheckTransaction(true);
  end;
end;

procedure TfrxIB_Query.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FDatabase) then
    SetDatabase(nil);
end;

function TfrxIB_Query.GetSQL: TStrings;
begin
	Result := FQuery.Sql;
end;

procedure TfrxIB_Query.SetSQL(Value: TStrings);
begin
	FQuery.Sql := Value;
end;

(** @see TfrxCustomDataset.SetUserName *)
procedure TfrxIB_Query.SetUserName(const Value: String);
begin
  inherited;
  UpdateDataTree;
end;

(** @see TfrxCustomDataset.InternalSetMaster *)
procedure TfrxIB_Query.InternalSetMaster(const Value: TfrxIBODataset);
begin
  FMaster := Value;
  if FMaster <> nil then
    FDataSource.DataSet := FMaster.GetDataSet
  else
    FDataSource.DataSet := nil;
end;

(** @see TfrxCustomDataset.GetActive *)
function TfrxIB_Query.GetActive: Boolean;
begin
  Result := Dataset.Active;
end;

(** @see TfrxCustomDataset.GetActive *)
procedure TfrxIB_Query.SetActive(Value: Boolean);
begin
  Dataset.Active := Value;
end;

(** @see TfrxCustomDataset.SetParent *)
procedure TfrxIB_Query.SetParent(AParent: TfrxComponent);
begin
  inherited;
  if (AParent <> nil) and (Report <> nil) then
  begin
    if IsDesigning and (Report.DataSets.Find(Self) = nil) then
    begin
      Report.DataSets.Add(Self);
      UpdateDataTree;
    end;
  end;
end;

(** @see TfrxCustomDataset.GetFiltered *)
function TfrxIB_Query.GetFiltered: Boolean;
begin
  Result := FQuery.Filtered;
end;

(** @see TfrxCustomDataset.GetFilter *)
function TfrxIB_Query.GetFilter: String;
begin
  Result := FQuery.Filter;
end;

(** @see TfrxCustomDataset.SetFiltered *)
procedure TfrxIB_Query.SetFiltered(Value: Boolean);
begin
  FQuery.Filtered := Value;
end;

(** @see TfrxCustomDataset.SetFilter *)
procedure TfrxIB_Query.SetFilter(const Value: String);
begin
  FQuery.Filter  := Value;
end;

(** @see TfrxCustomDataset.GetFields *)
function TfrxIB_Query.GetFields: TIB_Row;
begin
  Result := FQuery.Fields;
end;

(** @see TfrxCustomQuery.SetParams *)
procedure TfrxIB_Query.SetParams(Value: TfrxIB_Params);
begin
  FParams.Assign(Value);
end;

(** @see TfrxCustomQuery.ReadData *)
procedure TfrxIB_Query.ReadData(Reader: TReader);
begin
  frxReadCollection(FParams, Reader, Self);
//  UpdateParams;
end;

(** @see TfrxCustomQuery.WriteData *)
procedure TfrxIB_Query.WriteData(Writer: TWriter);
begin
  frxWriteCollection(FParams, Writer, Self);
end;

(** @see TfrxCustomQuery.DefineProperties *)
procedure TfrxIB_Query.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty('Parameters', ReadData, WriteData, True);
end;

(** @see TfrxCustomQuery.ParamByName *)
function TfrxIB_Query.ParamByName(const Value: String): TfrxIB_ParamItem;
begin
  Result := FParams.Find(Value);
end;

(** Copy the content of frx Params into IB Params *)
procedure TfrxIB_Query.UpdateParams;
var
  i: integer;
  Item: TfrxIB_ParamItem;
begin
  for i := 0 to FQuery.Params.ColumnCount - 1 do
  begin
    if Params.IndexOf(FQuery.Params[i].FieldName) <> -1 then
    begin
      Item := Params.Find(FQuery.Params[i].FieldName);
      if Assigned(Item) then
      begin

        if Trim(Item.Expression) <> '' then
        begin
          if not (IsLoading or IsDesigning) then
          begin
            Report.CurObject := Name;
            Item.Value := Report.Calc(Item.Expression);
          end;
        end;
        
        if not VarIsEmpty(Item.Value) then
        try
          FQuery.Params[i].AsVariant := Item.Value;
        except
          FQuery.Params[i].AsString := Item.Value;
        end;
      end;
    end;
  end;
end;

(** get Parameters *)
procedure TfrxIB_Query.DoAfterPrepare(Sender: TIB_Statement);
var
  i: integer;
  s: string;
  item: TfrxIB_ParamItem;
begin
  for i:=0 to FQuery.Params.ColumnCount-1 do
  begin
    s := FQuery.Params[i].FieldName;
    if Params.IndexOf(s) < 0 then
    begin
      item := TfrxIB_ParamItem.Create(Params);
      item.Name := s;
    end;
  end;
end;

(** Assign values to Parameters before opening *)
procedure TfrxIB_Query.DoBeforeOpen(IB_Dataset: TIB_Dataset);
begin
  UpdateParams;
end;

function TfrxIB_Query.GetMasterLinks: TStrings;
begin
  Result := FQuery.MasterLinks;
end;

procedure TfrxIB_Query.SetMasterLinks(const Value: TStrings);
begin
  FQuery.MasterLinks.Assign(Value);
end;

function TfrxIB_Query.GetMasterParamLinks: TStrings;
begin
  Result := FQuery.MasterParamLinks;
end;

procedure TfrxIB_Query.SetMasterParamLinks(const Value: TStrings);
begin
  FQuery.MasterParamLinks.Assign(Value);
end;

function TfrxIB_Query.GetPrepared: boolean;
begin
  Result := FQuery.Prepared;
end;

procedure TfrxIB_Query.SetPrepared(const Value: boolean);
begin
  FQuery.Prepared := Value;
end;

procedure TfrxIB_Query.UpdateDataTree;
begin
  if (Report <> nil) and (Report.Designer <> nil) then
    Report.Designer.UpdateDataTree;
end;

function TfrxIB_Query.GetVersion: string;
begin
  Result := S_VERSION;
end;

(** for savety: the object inspector of FR has a bug, if a property contains only read method. *)
procedure TfrxIB_Query.SetVersion(const Value: string);
begin
// do nothing
end;

function TfrxIB_Query.GetTransaction: TIB_Transaction;
begin
  Result := FQuery.IB_Transaction;
end;

procedure TfrxIB_Query.SetTransaction(const Value: TIB_Transaction);
begin

end;

var
  BmpCat: TBitmap;

initialization
  BmpCat := Graphics.TBitmap.Create;
  BmpCat.LoadFromResourceName(hInstance, 'frxIB_');

  frxObjects.RegisterCategory('IB_', BmpCat, 'IB Components');
  frxObjects.RegisterObject1(TfrxIB_Database, nil, '', 'IB_', 0, 37);
  frxObjects.RegisterObject1(TfrxIB_Query, nil, '', 'IB_', 0, 39);

finalization
  frxObjects.UnRegister(TfrxIB_Database);
  frxObjects.UnRegister(TfrxIB_Query);

  BmpCat.Free;

end.
